# --------------------------------------------------------------------------
###
#
# Copyright (c) Ensim Corporation 2000, 2001   All Rights Reserved.
#
# This software is furnished under a license and may be used and copied
# only  in  accordance  with  the  terms  of such  license and with the
# inclusion of the above copyright notice. This software or any other
# copies thereof may not be provided or otherwise made available to any
# other person. No title to and ownership of the software is hereby
# transferred.
#
# The information in this software is subject to change without notice
# and  should  not be  construed  as  a commitment by Ensim Corporation.
# Ensim assumes no responsibility for the use or  reliability  of its
# software on equipment which is not supplied by Ensim.
#
# $Id: DnsCommon.pm,v 1.24 2003/02/11 04:10:06 naris Exp $
# $Name:  $
# --------------------------------------------------------------------------

package DnsCommon;

use strict;

require Exporter;
our @ISA = qw (Exporter);
our @EXPORT_OK = qw (getZoneConfFile getZoneFile validateHostname validateDomain validateNetwork ip2Arpa getResolver handleUpdate verifyHostname);

use Net::DNS::Resolver;
use Sys::Hostname;
use Socket;

use lib ($ENV{OCW_SVCPATH} or "/usr/lib/opcenter") . "/bind";
use lib ($ENV{OCW_SVCPATH} or "/usr/lib/opcenter") . "/cmdline_common";

use ERRORS;
use Carper;

use constant ZONE_DIR_PREFIX  => "/var/named";
use constant ZONE_CONF_PREFIX => "/etc/bind";

use constant ERR_LONG_MAP => { NOERROR => "",
			       FORMERR  => "Internal Error",
			       SERVFAIL => "Internal Error",
			       NXDOMAIN => "Domain or Resource Record required is missing",
			       NOTIMP   => "Internal Error",
			       REFUSED  => "Operation refused",
			       YXDOMAIN => "Domain or Resource Record exists already",
			       YXRRSET  => "Domain or Resource Record exists already",
			       NXRRSET  => "Domain or Resource Record required is missing",
			       NOTAUTH  => "This name server is not authoritative for the given zone",
			       NOTZONE  => "Inappropriate Zone"};


use constant ERR_CODE_MAP => { NOERROR  => 0,
			       FORMERR  => 1,
			       SERVFAIL => 2,
			       NXDOMAIN => 3,
			       NOTIMP   => 4,
			       REFUSED  => 5,
			       YXDOMAIN => 6,
			       YXRRSET  => 7,
			       NXRRSET  => 8,
			       NOTAUTH  => 9,
			       NOTZONE  => 10};

use constant INT_32 => 4294967296;

sub getZoneConfFile ($) {
    my $zone = shift;
    if(validateNetwork($zone)) {
        return ZONE_CONF_PREFIX . "/zone." . ip2Arpa($zone);
    } else {
        return ZONE_CONF_PREFIX . "/zone.$zone";
    }
}

sub getZoneFile ($){
    my $zone = shift;
    if(validateNetwork($zone)) {
        return ZONE_DIR_PREFIX . "/db." . ip2Arpa($zone);
    } else {
        return ZONE_DIR_PREFIX . "/db.$zone";
    }
}

sub validateHostname($) {
    my $hostname = shift;

    # according to RFC 1123
    if (length $hostname > 255) {
        return '';
    }

    return ($hostname =~ m/^[a-zA-Z0-9]([a-zA-Z0-9\-]*[a-zA-Z0-9])?(\.[a-zA-Z0-9]([a-zA-Z0-9\-]*[a-zA-Z0-9])?)*$/);
}

sub validateDomain ($) {
    my $domain = shift;
    
    # according to RFC 1123
    return validateHostname ($domain);
}

sub validateNetwork ($) {
    my $network = shift;
    
    # do not enforce 0-255 check as spcified by RFC 952
    return ($network =~ m/^[0-9]+(\.[0-9]+)*$/);
}

sub validateIp ($) {
    my $ip = shift;
    return ($ip =~ /^\d+\.\d+\.\d+\.\d+$/);
}

# ip_to_arpa(address)
# Converts an IP address like 1.2.3.4 to 4.3.2.1.in-addr.arpa
sub ip2Arpa ($) {
    if ($_[0] =~ /^([\d\.]+)$/) {
        return join('.',reverse(split(/\./,$1))).".in-addr.arpa";
    }
    return $_[0];
}

sub getResolver () {
    my $res = new Net::DNS::Resolver;
    # assume this WP machine is running the DNS

    # for PR: 10416, changed 'localhost' to '127.0.0.1' in the
    # case that localhost does not resolve.
    $res->nameservers("127.0.0.1");
    $res->defnames(0);
    $res->recurse(0);

    return $res;
}

sub getDefaultRRTTL () {
    return 86400;
}

sub incSerial ($) {
    my $serial = shift;
    
    my $newSerial = ($serial + 1) % INT_32;

    # serial number for SOA better not be zero
    ++$newSerial if (!$newSerial);

    return $newSerial;
}

###
# This method would verify:
# - if the hostname specified actually has an A record
# - that A record maps to one ip address on a interface of this WP box.
# if second argument is supplied, second test is not performed
###
sub verifyHostname($;$) {
  # function adopted from MasterName.pm, masterName here really
  # means "hostname to queried"
  my $masterName = shift;
  my $newIP = shift;

  # verify the ip address of the new name is associated
  # with one of the inerfaces on this host
  my $res = new Net::DNS::Resolver;
  $res->defnames(0);
  my $query = $res->query ($masterName);

  # put all address into a hash
  my %newAddrHash = ();
  if ($query) {
    foreach my $rr ($query->answer) {
      if ($rr->type ne "A") {
	  next;
      }

      if ($rr->name eq $masterName) {
	  if ((!$newIP) || $newIP eq $rr->address) {
	      $newAddrHash{$rr->address} = 1;
	  } 
      }
    }
  } 

  if (! scalar(%newAddrHash)) {
        $cerr += $E_UNABLEFINDVALIDA and $cerr += { master => $masterName } ;
      return "Cannot find valid A record: $masterName" . ($newIP? "/$newIP." : "");
  }

  return if $newIP;

  my @hostIPList;

  for (split(/\n/, `/sbin/ip addr`)) {
      next unless /^\s*inet (\d+\.\d+\.\d+\.\d+)\//;
      return if $newAddrHash{$1};
  }

  $cerr += $E_NOITFMAP and $cerr += { name => $masterName } ;
  return "Name $masterName does not map to any interface on this host";
}


############### SHARED SUB FOR BINS ############################

###
# check if hostname conform to RFC 1123, die with an error
# message otherwise
###
sub dieHostname ($;$) {
    my $hostname = shift;
    my $disableEmpty = shift;

    if (!$disableEmpty) {
	return if $hostname eq '@';
    }

    $hostname =~ s/^(.*)\.$/$1/;

    $cerr += $E_MALFORMEDHOSTNAME and 
        $cerr += { hostname => $hostname } and 
        # die "Malformed hostname ($hostname)" 
        exit 1
        unless DnsCommon::validateHostname ($hostname);
}

sub dieWildHostname ($;$) {
    my $hostname = shift;
    my $disableEmpty = shift;

    if (!$disableEmpty) {
	return if $hostname eq '@';
    }

    return if $hostname eq '*';
    $hostname =~ s/^\*\.(.*)$/$1/;

    $hostname =~ s/^(.*)\.$/$1/;

    $cerr += $E_MALFORMEDHOSTNAME and 
        $cerr += { hostname => $hostname } and 
        die "Malformed hostname ($hostname)" 
        unless DnsCommon::validateHostname ($hostname);
}

sub dieIpAddr ($) {
    my $ipaddr = shift;

    $cerr += $E_INVALIDIP and $cerr += { ip => $ipaddr } and 
    die "Invalid IP address ($ipaddr)\n" 
        unless DnsCommon::validateIp ($ipaddr);
}

sub dieInZone ($$) {
    my $zone = shift;
    my $hostname = shift;

    return if $hostname eq '@';       # zone must be in itself
    return if $hostname =~ /[^\.]$/;  # relative hostname

    unless ( index ($hostname, "$zone.") != -1 ) {
        $cerr += $E_HOSTNOTINZONE and 
        $cerr += { host => $hostname , zone => $zone } ;
        die "Hostname ($hostname) is not in given zone ($zone)\n";
    }
}

###
# This method would return an integer mapping of the ERROR code.
###
sub getErrorCodeByStr ($) {
    return ERR_CODE_MAP->{$_[0]};
}

###
# This method would return a string describing the error.
###
sub getLongErrorByStr ($) {
    $cerr += ERR_LONG_MAP->{$_[0]} ;
    return ERR_LONG_MAP->{$_[0]};
}

###
# Handler for error. This method would not return
###
sub handleError ($) {
    my $err = shift;

    my $errCode = ERR_CODE_MAP->{$err};

    if (defined ($errCode)) {
	my $errStr = ERR_LONG_MAP->{$err};
    $cerr += $errStr ;
	print STDERR $errStr . "\n";
	
	exit $errCode;
    } else {
	print STDERR $err . "\n";
	exit 0x81;
    }
}

###
# This is the common method to send a update request to the name server.
# This routine would not return.
###
sub handleUpdate ($) {
    my $updater = shift;

    my $reply = $updater->send();

    if (!$reply) {
	print STDERR $updater->errorstring . "\n";
	exit 0x81
    }

    # handle error before connection
    unless (ref $reply) {
        print STDERR $reply . "\n";
        exit 0x81;
    }

    my $errCode = $reply->status;
    exit 0 if ($errCode eq "NOERROR"); # no error

    # handle error status by the name server
    $cerr += ERR_LONG_MAP->{$errCode} ;
    print STDERR ERR_LONG_MAP->{$errCode} . "\n";
    exit (ERR_CODE_MAP->{$errCode});
}

1;
