# --------------------------------------------------------------------------
###
#
# Copyright (c) Ensim Corporation 2000, 2001   All Rights Reserved.
#
# This software is furnished under a license and may be used and copied
# only  in  accordance  with  the  terms  of such  license and with the
# inclusion of the above copyright notice. This software or any other
# copies thereof may not be provided or otherwise made available to any
# other person. No title to and ownership of the software is hereby
# transferred.
#
# The information in this software is subject to change without notice
# and  should  not be  construed  as  a commitment by Ensim Corporation.
# Ensim assumes no responsibility for the use or  reliability  of its
# software on equipment which is not supplied by Ensim.
#
# $Id: VirtualDns.pm,v 1.11 2003/03/14 06:37:05 naris Exp $
# $Name:  $
# --------------------------------------------------------------------------

###
# This module provides an interface for Virutal DNS settings
###

package VirtualDns;

use strict;

use lib ($ENV{'OCW_SVCPATH'} || "/usr/lib/opcenter")."/bind";
use lib ($ENV{'OCW_SVCPATH'} || "/usr/lib/opcenter")."/cmdline_common";

require Exporter;
our @ISA = qw (Exporter);
our @EXPORT_OK = qw (getVirtual setVirtual delVirtual);

use IO::File;

use Sys::Hostname;

use DnsCommon;
use BindOptions qw (addListenOnIPAddr removeListenOnIPAddr);

use SymLock;
use LogConfFile;

use constant VIRTUAL_FILE => "/var/named/virtualDNS";
use constant VIRT_DNS_LOCK_NAME => "virt_dns.lock";

use ERRORS;
use Carper;

###
#
# This method returns a list of 2 elements:
# - name of the virtual DNS, 
# - IP address of the virtual DNS
#
# if virtual DNS is not configured, then these 2 elements would
# be empty.
#
###
sub getVirtual () {
    my $virtName = "";
    my $virtIP   = "";

    if (-r VIRTUAL_FILE ) {
        my $virtualdns;

        open (VIRTUALDNS, VIRTUAL_FILE) or (
            $cerr += $E_UNABLEREADFILE and 
            $cerr += { file => VIRTUAL_FILE } and 
            $cerr += "Unable to read file" and 
            $cerr += { file => VIRTUAL_FILE } and 
            die "Could not read " . VIRTUAL_FILE ) ;

        $virtualdns=<VIRTUALDNS>;
        close(VIRTUALDNS);

        chomp $virtualdns;
        
        if ($virtualdns) {
            $virtualdns =~ s/^\s*(.*)$/$1/g;
            
            ($virtName, $virtIP) = split /\s/, $virtualdns;
        }
    } 
    
    return (wantarray ? ($virtName, $virtIP) : $virtName);
}


###
#
# This method disable virtual DNS.
# The DNS server is restarted (if necessary) as a side effect.
# On success, it return false. Otherwise, an error string.
#
###
sub delVirtual () {
  SymLock::lock (VIRT_DNS_LOCK_NAME);
    my ($dnsoff) = _isServerOff();
    
    if (!$dnsoff) { 
        system("/etc/rc.d/init.d/named stop"); 
    }

    my $err="";
    
    # handle options update
    my @virtInfo = getVirtual();
    if ($err = BindOptions::removeListenOnIPAddr($virtInfo[1])) {
      SymLock::unlock (VIRT_DNS_LOCK_NAME);
	return $err;
    }

    if (-e VIRTUAL_FILE && (! unlink(VIRTUAL_FILE))) {
        $cerr += $E_UNABLEDELETEVIRTUALDNS and $cerr += {file => VIRTUAL_FILE };
        $err = "Cannot delete virtual DNS file " . VIRTUAL_FILE;
    }
    
    if (!$dnsoff) { 
        system("/etc/rc.d/init.d/named start"); 
    }
    
  SymLock::unlock (VIRT_DNS_LOCK_NAME);
    return $err;
}

###
#
# This method would set the value for virtual DNS. The DNS server
# is restarted (if necessary) as a side effect.
#
###
sub setVirtual ($$) {
    my $virtualname = shift;
    my $virtualIP   = shift;

    my @warning = ();

    if (! DnsCommon::validateHostname ($virtualname)) {
        $cerr += $E_MALFORMEDHOSTNAME and $cerr += { hostname => $virtualname} ;
        die "Malformed hostname: $virtualname\n";
    }

    if (! DnsCommon::validateIp ($virtualIP)) {
        $cerr += $E_MALFORMEDIP and $cerr += { ip => $virtualIP } ;
        die "Malformed IP address: $virtualIP\n";
    }

    # mansoor: we should not allow a virtual dns server to be added, which has
    # an IP that has already been aliased to eth0. because if we do allow such
    # a thing, then the delete virtual dns will remove that Ip from the list of
    # existing aliases on eth0

    my $existingIP = _getInterfaces();
    # my $existingIP = _getIPmap ();
    $cerr += $E_ALIASEXISTS and $cerr += { ip => $virtualIP } and exit 1
    if grep /^$virtualIP$/, @{$existingIP} ;

    my $cleanSub = sub { BindOptions::removeListenOnIPAddr ($virtualIP);
		       SymLock::unlock (VIRT_DNS_LOCK_NAME);};

  SymLock::lock (VIRT_DNS_LOCK_NAME);

    my ($dnsoff) = _isServerOff();

#    turned off.  PR 22491
#    if (!$dnsoff) {
#      my $err = DnsCommon::verifyHostname($virtualname, $virtualIP);
#      if ($err) {
#	  &$cleanSub;
#	  die "$err\n";
#      }
#    }

    my($hostname) = Sys::Hostname::hostname();
    my($hostIP, $hostnameIP);
    $hostnameIP = gethostbyname ($hostname);
    $hostIP = join (".",(unpack('C4',$hostnameIP)));
    if ($virtualIP eq $hostIP) {
	  &$cleanSub;
        $cerr += $E_VDNSSAMEASIP and $cerr += { vdns => $virtualIP } ;
        die "Virtual DNS server IP address $virtualIP cannot be the same with the main IP address of the server.\n";
    }
    

    BindOptions::addListenOnIPAddr ($virtualIP);

    open (VIRTUALDNS, "> " . VIRTUAL_FILE) or
        $cerr += $E_UNABLEWRITEFILE and $cerr += { file => VIRTUAL_FILE } and
      die "Could not write " . VIRTUAL_FILE . "  .\n";
    print VIRTUALDNS "$virtualname $virtualIP";
    close(VIRTUALDNS);

    # if dns is on, turn it off
    if (!$dnsoff) { 
        my $err = `/etc/rc.d/init.d/named_vwh_init restart 2>&1`;
	
	if ($?) {
	  SymLock::unlock (VIRT_DNS_LOCK_NAME);
	    die "$err\n";
	}
      SymLock::unlock (VIRT_DNS_LOCK_NAME);
    }

    if (scalar @warning) {
	return \@warning;
    } else {
	return ;
    }
}

sub _isServerOff () {
    my ($dnsoff) = system("/sbin/pidof named >/dev/null 2>&1");
    
    return $dnsoff;
}

sub _getInterfaces () {
    my @ipAddrs = ();

    open IP_ADDR_SHOW, "/sbin/ip addr show | ";

    my $buf;
    while ($buf = <IP_ADDR_SHOW>) {
	next if $buf =~ /^\d+/;
	next if $buf =~ /^\s*link/;
	next if $buf =~ /127\.0\.0\.1/;

	$buf =~ /\s*inet\s*([\d\.]*)/g;
	push @ipAddrs, $1;

    }

    my $err = close IP_ADDR_SHOW;

    if (!$err) {
	return;
    }
    
    return \@ipAddrs;
}

sub _getIPmap () {
    # list all IP addresses from /etc/virtualhosting/mappings/ipmap

    my @site_ips = ();
    open SITE_IPS, "/sbin/listaliases |";

    my $buf ;
    while ($buf = <SITE_IPS>) {
        chomp $buf;
        $buf =~ s/\s+$// ;
        push @site_ips, $buf ;
    }

    close SITE_IPS;
    return \@site_ips;
}






