# --------------------------------------------------------------------------
###
#
# Copyright (c) Ensim Corporation 2000, 2001   All Rights Reserved.
#
# This software is furnished under a license and may be used and copied
# only  in  accordance  with  the  terms  of such  license and with the
# inclusion of the above copyright notice. This software or any other
# copies thereof may not be provided or otherwise made available to any
# other person. No title to and ownership of the software is hereby
# transferred.
#
# The information in this software is subject to change without notice
# and  should  not be  construed  as  a commitment by Ensim Corporation.
# Ensim assumes no responsibility for the use or  reliability  of its
# software on equipment which is not supplied by Ensim.
#
# $Id: dns.pm,v 1.57 2003/02/25 11:46:19 sshivakumar Exp $
# $Name:  $
# --------------------------------------------------------------------------

package dns;

use strict;

require Exporter;
our @ISA = qw (Exporter);
our @EXPORT_OK = qw (create_zone delete_zone getZones renameZone);

use IO::File;
use Sys::Hostname;

use File::Basename;

use lib ($ENV{'OCW_SVCPATH'} || "/usr/lib/opcenter")."/bind";
use lib ($ENV{'OCW_SVCPATH'} || "/usr/lib/opcenter")."/cmdline_common";

use DnsCommon qw (getZoneConfFile getZoneFile validateHostname validateDomain validateNetwork ip2Arpa);
use MasterName qw (getMasterHostname);

use VirtualDns;
use DnsSoaDefaults;
use ZoneParser;
use SlaveServers;
use Zone_TSIG;

use SymLock;
use LogConfFile;

use constant LOCK_NAME => "bind.zone";

use constant RNDC_BIN => "/usr/sbin/rndc";

use constant ZONE_DIR_PREFIX  => "/var/named";
use constant ZONE_CONF_PREFIX => "/etc/bind";

use constant ZONE_CONF_TMPL => "named_conf_zone.tmpl";
use constant ZONE_FILE_TMPL => "zone_file.tmpl";

use constant BIND_CONF => "/etc/named.conf";
use constant BIND_LOG => "/var/log/messages";
use constant WP_BIND_CONF => "/etc/bind/bind.conf.wp";

use constant ERR_DIR_NOT_EXIST => 0x80;

use constant BIND_PROG_NAME => "named";


use ERRORS;
use Carper;

#Included to solve PR 18014 and PR 15351
use Socket;

###
# This method would use the default SOA setting to create a zone
# with a given name.
# 
# args:
# - zone name (either in domain name, or network number)
#
# return:
# - "" if OK, otherwise, error message
###
sub create_zone($$) {
    my $domain = shift;
    my $checkZoneExists = shift;

    my %zoneInfo = ();

    my $zoneSoaRef = DnsSoaDefaults::getValues();
    my $zone = "";
    my $zonefile;

    my $err;

    if (validateNetwork ($domain)) {
        # it is a network number
        $zone = ip2Arpa ($domain);
    } elsif (validateDomain ($domain)) {
        # it is a domain name
        $zone = "$domain";
    } else {
        $cerr += $E_MALFORMEDDOMAINORNW ;
        return "Malformed domain name or network number";
    }

    # start locking here
  SymLock::lock (LOCK_NAME);

    # check if the zone exists already on this webppliance
    if ($checkZoneExists) {
	my $resolver = DnsCommon::getResolver();
	my $zoneSOA = $resolver->query ($zone, "SOA");
	if (defined $zoneSOA) {
	    my $wpMName = getMasterHostname();

	    if (($zoneSOA->answer)[0]->mname eq $wpMName) {
	      SymLock::unlock (LOCK_NAME);
		return 'YXDOMAIN';
	    }
        return 'YXDOMAIN';
	} 

	# we want to be sure that even if BIND is not running
	# we don't add in extra line to bind.conf.wp
	my $pattern = _getZonePattern ($zone);
	my $inputHandle = new IO::File (WP_BIND_CONF, "r");

	if (_skipToLine ($inputHandle, $pattern)) {
	  SymLock::unlock (LOCK_NAME);
	    return 'YXDOMAIN';
	}
    }
    
    # make sure all dirs exists
    _assertDir (ZONE_DIR_PREFIX);

    $zonefile = getZoneFile ($zone);

    # setup up zone info
    $zoneInfo{zone}     = $zone;
    $zoneInfo{zonefile} = $zonefile;
    $zoneInfo{update_key_id} = Zone_TSIG::getWPDefaultKeyID();


#Included to solve PR 18014 and PR 15351
    my $zoneslaves = "";
    my $iaddr;
    my $iaddr_str;
    # add additional NS record for default slave servers
    my $slaveserverList = SlaveServers::getSlaves;
    if (scalar @{$slaveserverList}) {
        foreach my $slaveNS (@{$slaveserverList}) {
	    if($slaveNS ne "") {
		$iaddr = gethostbyname($slaveNS);
		if($iaddr) {
#inet_ntoa expects the size if IN parameter to be 4 bytes
		    $iaddr_str = inet_ntoa($iaddr);
		    $zoneslaves .= $iaddr_str . "; ";
		}
	    }
	}
    }
    $zoneInfo{slaveserver} = $zoneslaves;


    # setup SOA
    $zoneSoaRef->{master} = getMasterHostname();
    $zoneSoaRef->{email} or $zoneSoaRef->{email} = "root\@$zoneSoaRef->{master}";
    my $y = `date +%Y`;
    chomp($y);
    my $m = `date +%m`;
    chomp($m);
    my $d = `date +%d`;
    chomp($d);
    my $date = $y . $m . $d . "01";
    $zoneSoaRef->{serial} = sprintf("%d",$date);
    #$zoneSoaRef->{serial} = 1;

    # change the email address format to that used by BIND
    $zoneSoaRef->{email} =~ tr/@/\./;
    $zoneSoaRef->{email} =~ s/^(.+[^.])\.?$/$1\./g;

    $zoneSoaRef->{master} =~ s/^(.+[^.])\.?$/$1\./g;

    # generate config of the zone
    my $zoneConfFile = getZoneConfFile($zone);
    _assertDir (dirname($zoneConfFile));

    my $zoneConfFH = new IO::File $zoneConfFile, "w";
    _expandTmpl (($ENV{'OCW_SVCPATH'} || "/usr/lib/opcenter") . "/bind/" . 
                 ZONE_CONF_TMPL, 
                 \%zoneInfo, $zoneConfFH);
    if (!$zoneConfFH->close()) {
      SymLock::unlock (LOCK_NAME);
      $cerr += $E_UNABLEGENERATEZONECFG ;
	return "Failed to generate zone config file" ;
    }

    # generate zone file
    my $zoneFH = new IO::File "$zonefile", "w";
    _expandTmpl (($ENV{'OCW_SVCPATH'} || "/usr/lib/opcenter") . "/bind/" . 
                 ZONE_FILE_TMPL, 
                 $zoneSoaRef, $zoneFH);

    # if virtual DNS is on, add an extra NS RR
    my @extraRRList = ();

    my $virtHost;
    if ($virtHost = VirtualDns::getVirtual()) {
      push @extraRRList, {name => '@', type => 'NS', rdata => "$virtHost."};
    }

    # add additional NS record for default slave servers
    my $slaveList = SlaveServers::getSlaves;
    if (scalar @{$slaveList}) {
	foreach my $slaveNS (@{$slaveList}) {
	    push @extraRRList, {name => '@', type => 'NS', rdata => "$slaveNS."};
	}
    }
    
    _insertRecord($zoneFH, \@extraRRList);

    if (!$zoneFH->close()) {
      SymLock::unlock (LOCK_NAME);
      $cerr += $E_UNABLEGENERATEZONEFILE ;
	return "Failed to generate zone file";
    }

    # add an 'include' directive to /etc/named.conf
    if ($err = _namedConfAddInclude ($zoneConfFile)) {
      SymLock::unlock (LOCK_NAME);
	return $err;
    }

  SymLock::unlock (LOCK_NAME);

    if ($err = _reloadBIND()) {
	return $err;
    }

    # double check that BIND is running
    if (!_isBINDRunning()) {
        $cerr += $E_BINDRELOADFAIL  and $cerr += { file => BIND_LOG } ;
	return "BIND reload failed. Please consult " . BIND_LOG;
    }
}

sub delete_zone($) {
    my $zone = shift;
    my $inputHandle;
    my $outputHandle;
    my $err;

    my $namedConfDir  = dirname WP_BIND_CONF;
    my $namedConfFile = basename WP_BIND_CONF;

    # start locking here
  SymLock::lock (LOCK_NAME);

    if (!($inputHandle = new IO::File WP_BIND_CONF, "r")) {
      SymLock::unlock (LOCK_NAME);
        $cerr += $E_UNABLEOPENFILE and $cerr += { file => WP_BIND_CONF } ;
        return "Unable to open " . WP_BIND_CONF
    }

    if (!($outputHandle = LogConfFile::open_conf ($namedConfDir, 
						  "/var/log/conf",
                                                  $namedConfFile, "w"))) {
      SymLock::unlock (LOCK_NAME);
        $cerr += $E_UNABLEUPDATEFILE and $cerr += { file => WP_BIND_CONF };
        return "Unable to update " . WP_BIND_CONF;
    }

    my $pattern = _getZonePattern ($zone);
    
    # we just want to remove the line from /etc/bind/bind.conf.wp
    while ($_ = _skipToLine ($inputHandle, $pattern, $outputHandle)) {
    }
    
    $inputHandle->close();
    if (! LogConfFile::close_conf ($outputHandle,  $namedConfDir, 
				   "/var/log/conf", $namedConfFile)) {
      SymLock::unlock (LOCK_NAME);
        $cerr += $E_UNABLEUPDATEFILE and $cerr += { file => WP_BIND_CONF };
        return "Unable to update " . WP_BIND_CONF;
    }

    # need to remove the jounal file (if exists)
    _cleanupFile ($zone);

    if ($err = _reloadBIND()) {
      SymLock::unlock (LOCK_NAME);
	return $err;
    }

    # double check that BIND is running
    if (!_isBINDRunning()) {
      SymLock::unlock (LOCK_NAME);
        $cerr += $E_BINDRELOADFAIL  and $cerr += { file => BIND_LOG } ;
	return "BIND reload failed. Please consult " . BIND_LOG;
    }

    SymLock::unlock (LOCK_NAME);
    return $err
}


###
# This method returns a reference to hash of zone info
# keyed by zone name
###
sub getZones () {
    my $parser = new ZoneParser;
    $parser->parse_file(BIND_CONF);
    
    return $parser->zones;
}

###
# This method would rename a zone. All the RR of a zone is transferred.
#
# args:
#  - oldZoneame
#  - newZoneName
#
# return:
#  - false if ok, otherwise, error
###
sub renameZone ($$) {
    my $oldZone = shift;
    my $newZone = shift;
    
    my $pattern = qr/\Q$oldZone\E (?= \. (?! [a-zA-Z0-9\-]) | [^a-zA-Z0-9\-]) /ix;
    my $err;
    
    # start locking here
  SymLock::lock (LOCK_NAME);
    print "acquired lock\n";

    # check if the target zone exists first
    my $zonesRef = getZones();
    if ($zonesRef->{$newZone}) {
      SymLock::unlock (LOCK_NAME);
	return "target zone exists already: $newZone";
    }
    
    if (!$zonesRef->{$oldZone}) {
      SymLock::unlock (LOCK_NAME);
	return "zone does not exist: $oldZone";
    }

    # first, we need to stop BIND via rndc stop
    if ($err = _stopBIND()) {
      SymLock::unlock (LOCK_NAME);
        return $err;
    }
    
    # move the zone file
    my $oldZoneFile = getZoneFile ($oldZone);
    my $newZoneFile = getZoneFile ($newZone);

    if (!_filterFile ($oldZoneFile, $newZoneFile,
                      $pattern, $newZone)) {
      SymLock::unlock (LOCK_NAME);
        return "Failed to update Zone File";
    }


    # move the include file
    my $oldZoneConfFile = getZoneConfFile ($oldZone);
    my $newZoneConfFile = getZoneConfFile ($newZone);
    
    if (!_filterFile ($oldZoneConfFile, $newZoneConfFile,
                      $pattern, $newZone)) {
      SymLock::unlock (LOCK_NAME);
        return "Failed to update Zone File";
    }


    # update named.conf
    my @namedConfErr = ();

    if ($err = delete_zone ($oldZone)) {
	push @namedConfErr, $err;
    }

    if ($err = _namedConfAddInclude ($newZoneConfFile)) {
	push @namedConfErr, $err;
    }

    if (scalar @namedConfErr) {
	$err = join "\n", @namedConfErr;
    }

    # start BIND again
    $err = _startBIND();

    # double check if BIND is started
    if (!_isBINDRunning()) {
      SymLock::unlock (LOCK_NAME);
      $cerr += $E_BINDSTARTUPFAIL and $cerr += { file => BIND_LOG } ;
	return "BIND startup failed. Please consult " . BIND_LOG;
    }

    unlink $oldZoneFile;
    unlink $oldZoneConfFile;

  SymLock::unlock (LOCK_NAME);
    return $err;
}

############################# Helper subs #####################################

sub _stopBIND() {
    my $cmd = "/etc/rc.d/init.d/named stoprndc";

    print $cmd . "\n";
    my $err = `$cmd 2>&1`;

    if ($?) {
	chomp $err;
	return $err;
    }

    return;
}

sub _startBIND() {
    my $cmd = "/etc/rc.d/init.d/named start";

    my $err = `$cmd 2>&1`;

    if ($?) {
	chomp $err;
	return $err;
    }

    return;
}

sub _reloadBIND() {
    my $cmd = RNDC_BIN . " reload";

    return if (!_isBINDRunning());

    my $err = `$cmd 2>&1`;

    if ($?) {
	chomp $err;
	return $err;
    }

    return;
}

sub _isBINDRunning() {
    my @result = `ps aux | grep named`;

    foreach my $line (@result) {
	if ($line !~ /grep/) {
	    return 1;
	}
    }

    return;
}

sub _filterFile ($$$$) {
    my $inputFile = shift;
    my $outputFile = shift;

    my $inputFH  = new IO::File "$inputFile", "r";
    my $outputFH = new IO::File "$outputFile", "w";
  
    _filterHandle ($inputFH, $outputFH, $_[0], $_[1]);

    $inputFH->close();
    return $outputFH->close();
}

sub _filterHandle ($$$$) {
    my $inputFH = shift;
    my $outputFH = shift;
    my $pattern = shift;
    my $sub = shift;

    my $buf;
    while ($buf = <$inputFH>) {
        $buf =~ s/$pattern/$sub/;
        print $outputFH $buf;
    }
}

sub _cleanupFile ($) {
    my $zone = shift;

    my $zoneConfFile = getZoneConfFile($zone);
    my $zoneFile="";

    my $parser = new ZoneParser;
    $parser->parse_file (BIND_CONF);
    $zoneFile = $parser->zones->{$zone}->{file}; 
    
    if (!$zoneFile) {
        $zoneFile = getZoneFile($zone);
    }
    
    my $jnlFile =  $zoneFile. ".jnl";

    if (-e $jnlFile) {
        unlink $jnlFile;
    }

    if (-e $zoneFile) {
	unlink $zoneFile;
    }

    if (-e $zoneConfFile) {
	unlink $zoneConfFile;
    }

    return "";
}

sub _namedConfAddInclude ($) {
    my $zoneConfFile = shift;
    
    my $namedConfDir  = dirname WP_BIND_CONF;
    my $namedConfFile = basename WP_BIND_CONF;

    my $namedConfFH = LogConfFile::open_conf ($namedConfDir, "/var/log/conf",
                                              $namedConfFile, "a");
    if (!$namedConfFH) {
        # it is not necessary to delete the zone files
        $cerr += $E_UNABLEOPENCFGFILE;
        return "unable to open config file for BIND";
    }

    print $namedConfFH "include \"$zoneConfFile\";\n";

    if (!LogConfFile::close_conf ($namedConfFH, $namedConfDir, 
                                  "/var/log/conf", $namedConfFile)) {
        $cerr += $E_UNABLEUPDATEFILE and $cerr += { file => WP_BIND_CONF } ;
        return "Unable to update " . WP_BIND_CONF;
    }
}

sub _assertDir ($) {
  my $dir = shift;

  return if ( -d $dir );
  return if ( -l $dir );

  exit ERR_DIR_NOT_EXIST;
}

sub _insertRecord($$) {
  my $outputFH = shift;
  my $rrListRef = shift;

  foreach my $rr (@{$rrListRef}) {
    print $outputFH <<END_RR
$rr->{name}                       IN $rr->{type}           $rr->{rdata}
END_RR

  }
}

sub _getZonePattern ($) {
    my $domain = shift;
    my $target = getZoneConfFile ($domain);
    my $pattern = qr/^\s*include\s*"$target";\s*$/;

    return $pattern;
}

sub _skipToLine ($$;$) {
    my $inputHandle = shift;
    my $pattern = shift;
    my $outputHandle = shift;
    
    my $buf;
    while ($buf = <$inputHandle>) {
        chomp $buf;
        if ($buf =~ m/$pattern/) {
            return $buf;
        }
        
        if ($outputHandle) {
            print $outputHandle $buf;
            print $outputHandle "\n";
        }
    }
}

sub _expandTmpl ($$;$) {
    my $tmplName = shift;
    my $in = shift;
    my $output_handle = shift;

    my @retBuf=();
    my $input_handle = new IO::File "$tmplName", "r";

    if (defined ($input_handle)) {
        my $buf;
        my $line="";
        
        while ($buf = <$input_handle>) {
            chomp $buf;
            # escape "
            $buf =~ s/(")/\\$1/g;

            # XXXX MIGHT WANT TO CHECK IF VAR(s) ARE DEFINED
            eval "\$line = \"$buf\";" ;
            
            if ($output_handle) {
                print $output_handle "$line\n";
            } else {
                push @retBuf, $line;
            }
        }
    }
    undef $input_handle;

    if (!$output_handle) {
        return \@retBuf;
    }
}

1;
